import {watch} from "vue";
import useExpr2Store from "@/store/expr2/expr2Store.ts";
import {storeToRefs} from "pinia";
import {Operation} from "@/store/expr2/defaultOperation.ts";

export enum Algorithm {
    FCFS,
    SSTF,
    SCAN_FirstDown,
    SCAN_FirstUp
}

export class Disk {
    constructor() {
        const {
            algorithm,
            operations,
            originOperations,
            startTrack,
            trackCount,
            failInfo,
            tableChanged,
            dataChanged
        } = storeToRefs(useExpr2Store());
        watch(tableChanged, (newVal) => {
            if(newVal === false)
                return;
            failInfo.value = "";
            switch (algorithm.value) {
                case Algorithm.FCFS:
                    operations.value = originOperations.value.map(x => x);
                    break;
                case Algorithm.SSTF:
                    operations.value = this.SSTF(originOperations.value.map(x => x), startTrack.value);
                    break;
                case Algorithm.SCAN_FirstDown:
                    operations.value = this.SCAN(originOperations.value.map(x => x), startTrack.value, false);
                    break;
                case Algorithm.SCAN_FirstUp:
                    operations.value = this.SCAN(originOperations.value.map(x => x), startTrack.value, true);
                    break;
            }
            for (let i = 0; i < operations.value.length; i++) {
                if (operations.value[i].track < 0 || operations.value[i].track > trackCount.value) {
                    failInfo.value = i + `: track(${operations.value[i].track}) out of range`;
                }
            }
            dataChanged.value = true;
            tableChanged.value = false;
        });
    }

    SSTF(operations: Operation[], startTrack: number): Operation[] {
        let ans: Operation[] = [];
        operations.sort((a, b) => a.track - b.track);
        let set: Set<number> = new Set();

        function FindTrack(): number {
            let track = startTrack;
            if (track < operations[0].track)
                track = 0;
            else if (track > operations[operations.length - 1].track)
                track = operations.length - 1;
            else
                for (let i = 0; i < operations.length - 1; i++) {
                    if (operations[i].track === track) {
                        track = i;
                        break;
                    } else if (operations[i].track <= track && track <= operations[i + 1].track) {
                        track = track - operations[i].track < operations[i + 1].track - track ? i : i + 1;
                        break;
                    }
                }
            return track;
        }

        let track = FindTrack();
        ans.push(operations[track])
        set.add(track);
        //O(n)
        while (set.size < operations.length) {
            let candidate: number[] = [];

            let left = track, right = track;
            while (left >= 0) {
                if (set.has(left))
                    left--;
                else {
                    candidate.push(left);
                    break;
                }
            }
            while (right <= operations.length - 1) {
                if (set.has(right))
                    right++;
                else {
                    candidate.push(right);
                    break;
                }
            }
            let min = Number.MAX_VALUE;
            for (let t of candidate) {
                if (Math.abs(operations[t].track - operations[track].track) < min) {
                    min = Math.abs(operations[t].track - operations[track].track);
                    track = t;
                }
            }
            ans.push(operations[track]);
            set.add(track);
        }
        return ans;
    }

    SCAN(operations: Operation[], startTrack: number, direction: boolean): Operation[] {
        let ans: Operation[] = [];
        operations.sort((a, b) => a.track - b.track);
        let track;

        if(direction)
            track = operations.findIndex(x => x.track >= startTrack);
        else
            track = operations.length - 1 - operations.slice().reverse().findIndex(x => x.track <= startTrack);
        if(track === -1)
            track = operations.length - 1;
        if(track === operations.length)
            track = 0;
        startTrack = track;
        ans.push(operations[track]);
        while (track + (direction ? 1 : -1) >= 0 && track + (direction ? 1 : -1) < operations.length) {
            track += (direction) ? 1 : -1;
            ans.push(operations[track]);
        }
        track = startTrack;
        while (track + (direction ? -1 : 1) >= 0 && track + (direction ? -1 : 1) < operations.length) {
            track += (direction) ? -1 : 1;
            ans.push(operations[track]);
        }
        return ans;
    }
}